"""
The expiry date is in the YYMMDD format in the EU FMD regulated barcodes. The program validates the expiry date. 
MM to be between valid values of 01 and 12. 
DD to be between valid values of 00 and 31. Taking account of months with date upto 30 days and Feb with 28 & 29 days. 
The logic also allows 00 for DD, this refers to the last day of the month. 

"""


from datetime import datetime


def expiry_date_check(e:str):
    my_year = e[:2]
    my_month = e[2:4]
    my_date = e[4:]
    
    if (int(my_month) not in range(1,13)):
        return False
    elif (int(my_date) not in range(32)):
        return False
    
    if my_date == '00':
        if my_month == '02' and int(my_year) % 4 == 0:
            my_date = '29'
        elif my_month == '02' and int(my_year) % 4 != 0:
            my_date = '28'
        elif my_month in ['01', '03', '05', '07', '08', '10', '12']:
            my_date = '31'
        else:
            my_date = '30'
    
    if my_month == '02' and int(my_year) % 4 == 0 and my_date <= '29':
        pass
    elif my_month == '02' and int(my_year) % 4 != 0 and my_date <= '28':
        pass
    elif my_month in ['01', '03', '05', '07', '08', '10', '12'] and my_date <= '31':
        pass
    elif my_month in ['04', '06', '09', '11'] and my_date <= '30':
        pass
    else:
        return False

    actual_date = my_year + my_month + my_date
    if actual_date > datetime.today().strftime('%y%m%d'):
        return True
    else:
        return False
    


"""
The function parse the GS1 Datamatrix barcode used for medicine packs.
The barcode scanners in the default setup outputs the scan is in below format:
]d201034531200000111719112510ABCD1234<GS>2110EW354EWER
(01)03453120000011(17)191125(10)ABCD1234<GS>(21)10EW354EWER
(01) = GTIN Identifier - Fixed 14 chars
(17) = Expiry Date Identifer - Fixed 6 Chars
(10) = Batch Identifier - Variable length. (upto 20 chars usually between 4 - 10 Chars)
(21) = Serial Number - Variable length. (upto 20 chars - Usually between 12 - 20 Chars)
(710 or 711 or 712 or 713 or 714) = Optional NHRN identifiers.
The symbology identifier ]d2 and for the second FNC1, when used as a separator character is <GS> Group-Separator.

"""


from .gs1_gtin_validation import gtin_check #The GTIN validation module
from .expiry_date_validation import expiry_date_check #The Expiry validation module

result = {'SCHEME': 'GTIN'}


def gs1_gtin(barcode: str) -> dict: 
    
    if chr(29) in barcode:

        while barcode:

            if barcode[:2] == '01':
                result['GTIN'] = barcode[2:16]
                if len(barcode) > 16:
                    barcode = barcode[16:]
                else:
                    barcode = None

            elif barcode[:2] == '17':
                result['EXPIRY'] = barcode[2:8]
                if len(barcode) > 8:
                    barcode = barcode[8:]
                else:
                    barcode = None

            elif barcode[:2] == '10':
                if chr(29) in barcode:
                    for i, c in enumerate(barcode):
                        if ord(c) == 29:
                            result['BATCH'] = barcode[2:i]
                            barcode = barcode[i+1:]
                            break
                else:
                    result['BATCH'] = barcode[2:]
                    barcode = None
                    
            elif barcode[:2] == '21':
                if chr(29) in barcode:
                    #print("In the serial GS check")
                    for i, c in enumerate(barcode):
                        if ord(c) == 29:
                            result['SERIAL'] = barcode[2:i]
                            barcode = barcode[i+1:]
                            break
                else:
                    result['SERIAL'] = barcode[2:]
                    barcode = None


            elif barcode[:3] in ['710', '711', '712', '713', '714']:
                if chr(29) in barcode:
                    for i, c in enumerate(barcode):
                        if ord(c) == 29:
                            result['NHRN'] = barcode[3:i]
                            barcode = barcode[i+1:]
                            break
                else:
                    result['NHRN'] = barcode[3:]
                    barcode = None
            else:
                return {'ERROR': 'INVALID BARCODE'}
    else:
        return {'ERROR': 'INVALID BARCODE'}

    if 'GTIN' and 'BATCH' and 'EXPIRY' and 'SERIAL' in result.keys():
        if gtin_check(result['GTIN']) == False and expiry_date_check(result['EXPIRY']) == False:
            return {'ERROR': 'INVALID GTIN & EXPIRY DATE'}
        elif expiry_date_check(result['EXPIRY']) == False:
            return {'ERROR': 'INVALID EXPIRY DATE'}
        elif gtin_check(result['GTIN']) == False:
            return {'ERROR': 'INVALID GTIN'}
        else:
            return result
    else:
        return {'ERROR': 'INVALID BARCODE'}
    




"""
ASCII Code representation of a PPN code:
[)><30>06<29>9N 110375286414<29>1T12345ABCDE<29>D160617<29>S12345ABCDEF98765<30><4>

Actual Format including the ASCII chars is as below:

MACRO 06: The header [)> + ASCII 30 + ASCII 06 + ASCII 29 will be transmitted by the barcode reader before the data in the message and the trailer ASCII 30 + ASCII 4 will be transmitted afterwards.
OR
MACRO 05: The header [)> + ASCII 30 + ASCII 05 + ASCII 29 will be transmitted by the barcode reader before the data in the message and the trailer ASCII 30 + ASCII 4 will be transmitted afterwards. 

<29> is the GS separator. It’s used before the start of each identifier string. 

Identifier	Name	        Ex
9N	        PPN Code	    110375286414 (12 digits)
1T	        Batch	        12345-ABCDE (upto 20 digits)
D	        Expiry date	    160617 (YYMMDD)
S	        Serial number	12345ABCDEF98765 (between 12-20 chars)
8P	        GTIN	        08701234567896 (this is optional)

"""


from .expiry_date_validation import expiry_date_check #The Expiry validation module
from .ifa_ppn_validation import ppn_check

result = {'SCHEME': 'PPN'}

def ifa_ppn(barcode: str) -> dict: 

    while barcode:

        if barcode[:2] == '9N':
            result['PPN'] = barcode[2:14]
            if len(barcode) > 14:
                barcode = barcode[15:]
            else:
                barcode = None

        elif barcode[:1] == 'D':
            result['EXPIRY'] = barcode[1:7]
            if len(barcode) > 7:
                barcode = barcode[8:]
            else:
                barcode = None

        elif barcode[:2] == '1T':
            if chr(29) in barcode:
                for i, c in enumerate(barcode):
                    if ord(c) == 29:
                        result['BATCH'] = barcode[2:i]
                        barcode = barcode[i+1:]
                        break
            else:
                for i, c in enumerate(barcode):
                    if ord(c) == 30:
                        result['BATCH'] = barcode[2:i]
                        barcode = None
                
        elif barcode[:1] == 'S':
            if chr(29) in barcode:
                for i, c in enumerate(barcode):
                    if ord(c) == 29:
                        result['SERIAL'] = barcode[1:i]
                        barcode = barcode[i+1:]
                        break
            else:
                for i, c in enumerate(barcode):
                    if ord(c) == 30:
                        result['SERIAL'] = barcode[1:i]
                        barcode = None
        
        elif barcode[:2] == '8P':
            result['GTIN'] = barcode[2:16]
            if len(barcode) > 16:
                barcode = barcode[17:]
            else:
                barcode = None

        else:
            return {'ERROR': 'INVALID BARCODE'}

    if 'PPN' and 'BATCH' and 'EXPIRY' and 'SERIAL' in result.keys():
        if ppn_check(result['PPN']) == False and expiry_date_check(result['EXPIRY']) == False:
            return {'ERROR': 'INVALID PPN & EXPIRY DATE'}
        elif expiry_date_check(result['EXPIRY']) == False:
            return {'ERROR': 'INVALID EXPIRY DATE'}
        elif ppn_check(result['PPN']) == False:
            return {'ERROR': 'INVALID PPN'}
        else:
            return result
    else:
        return {'ERROR': 'INVALID BARCODE'}


"""
The function perform validation on the GS1-GTIN(14) based on the check digit detailed in the link below:
https://www.gs1.org/services/how-calculate-check-digit-manually

"""

def gtin_check(gtin: str):
    reverse_gtin = gtin[-2::-1] #Reversing the string without the check-digit
    digit_multipler3 = []
    digit_multipler1 = []

    for i, l in enumerate(reverse_gtin): 
        if i % 2 == 0:
            digit_multipler3.append(int(l))
        else:
            digit_multipler1.append(int(l))

    digit_sum = (sum(digit_multipler3) * 3) + sum(digit_multipler1)
    nearest_ten = round(digit_sum/10) * 10
    check_sum_digit = nearest_ten - digit_sum
    if check_sum_digit < 0:
        check_sum_digit = 10 + check_sum_digit
    return check_sum_digit == int(gtin[-1])

    



"""
This function contains code to identify the barcode type: GS1-GTIN or IFA-PPN.

"""

from .gs1_gtin import gs1_gtin
from .ifa_ppn import ifa_ppn

def pharma_datamatrix(barcode: str) -> dict:
    """
    The function takes barcode string and return parsed objects in a `dict`. 
    Optionally user can pass bool value to validate the barcode
    Parameters:
        barcode: string type. This is the barcode string with the <GS> seperators. 
     
    Returns:
        dict: Returns dictionary object with SCHEME, PPN, GTIN, EXPIRY, BATCH, SERIAL & NHRN as keys for valid requests 
        or relevant error strings.
    """
    if barcode[:3] == ']d2': #Most barcode scanners prepend ']d2' identifier for the GS1 datamatrix. This senction removes the identifier.
        barcode = barcode[3:]
        result = gs1_gtin(barcode)
    
    elif barcode[:2] in ['01', '21', '17', '10', '71']:
        result = gs1_gtin(barcode)

    elif barcode[:6] == ('[)>'+chr(30)+'06') or ('[)>'+chr(30)+'05'): # MACRO 06 or MACRO 05
        barcode = barcode[7:] # Removes the leading ASCII seperator after the scheme identifier.
        result = ifa_ppn(barcode)
   
    else:
        result = {'ERROR': 'INVALID BARCODE'}
        
    return result



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(name='pharma_datamatrix',
      version='1.3.1',
      description='Parse EU FMD compliant GS1 & PPN Datamatrix barcodes',
      packages=['pharma_datamatrix'],
      auther = 'Kuldeep Saxena',
      author_email = 'saxenak02@gmail.com',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://github.com/PinchofLogic/pharma_datamatrix",
      zip_safe=False)


"""
Each PPN requires two Modulo 97-calculated check digits for additional data integrity. 
To calculate the check digits, the ASCII value of the alphanumeric characters is used and multiplied 
by an ascending weight factor. The weighting of the digits starts on the left with two and increases 
by one for each following digit. The results of each multiplication are summed up and divided by 97, 
and the remainder are the check digits. If the remainder is only one digit, a leading zero is added.
"""

def ppn_check(ppn: str) -> bool:
    i = 0
    weight = 2
    digit_sum = 0
    for i in range(10):
        digit_sum += (ord(ppn[i]) * weight)
        weight += 1
    print(digit_sum)
    check_digit = digit_sum % 97
    return check_digit == int(ppn[-2:]) #PPN last two chars are converted to int to remove any leading zero.


